package com.ti.mobpo.ui.viewmodels

import androidx.lifecycle.ViewModel
import androidx.lifecycle.viewModelScope
import com.ti.mobpo.data.Favourite
import com.ti.mobpo.data.FavouritesRepository
import com.ti.mobpo.model.PokemonDetails
import com.ti.mobpo.model.PokemonSpecies
import com.ti.mobpo.network.PokeApi
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.flow.StateFlow
import kotlinx.coroutines.flow.asStateFlow
import kotlinx.coroutines.launch
import java.io.IOException


private const val SHOW_LIMIT = 20;
class PokeSearchViewModel(private val favouritesRepository: FavouritesRepository) : ViewModel() {
    private val service = PokeApi.retrofitService;

    private val _pokemonDetails = MutableStateFlow<List<PokemonDetails>?>(null)
    val pokemonDetails: StateFlow<List<PokemonDetails>?> = _pokemonDetails.asStateFlow()
    private val _initialPokemonList = MutableStateFlow<List<PokemonSpecies>?>(null)

    private val _isLoading = MutableStateFlow(false)
    val isLoading: StateFlow<Boolean> = _isLoading.asStateFlow()
    init {
        fetchPokemonSpecies()
    }

    private fun fetchPokemonSpecies() {
        viewModelScope.launch {
            try {
                val response = service.getPokemon()
                _initialPokemonList.value = response.results.sortedBy { it.name.lowercase() }
            } catch (e: IOException) {
                /*TODO*/
            }
        }
    }

    fun search(query: String) {
        _isLoading.value = true
        viewModelScope.launch {
            try {
                val filteredList = _initialPokemonList.value?.binarySearch(query)
                val detailsList = mutableListOf<PokemonDetails>()
                if (filteredList != null) {
                    for (index in filteredList until minOf(_initialPokemonList.value?.size ?: 0, filteredList + SHOW_LIMIT)) {
                        val pokemonSpecies = _initialPokemonList.value!![index]
                        val details = service.getPokemonDetails(extractPokemonId(pokemonSpecies.url))
                        val isFavorite = favouritesRepository.isFavourite(details.id)
                        detailsList.add(details.copy(isFavorite = isFavorite))
                    }
                }
                _pokemonDetails.value = detailsList
            } catch (e: IOException) {
                /* Handle error */
            } finally {
                _isLoading.value = false
            }
        }
    }

    private fun List<PokemonSpecies>.binarySearch(query: String): Int {
        var low = 0
        var high = this.size - 1
        while (low <= high) {
            val mid = (low + high) ushr 1
            val comparison = this[mid].name.compareTo(query, ignoreCase = true)
            when {
                comparison == 0 -> return mid
                comparison < 0 -> low = mid + 1
                else -> high = mid - 1
            }
        }
        return low
    }

    private fun extractPokemonId(url: String): Int {
        val parts = url.split("/")
        return parts[parts.size - 2].toInt()
    }

    fun toggleFavorite(pokemonId: Int) {
        _pokemonDetails.value = _pokemonDetails.value?.map { pokemon ->
            if (pokemon.id == pokemonId) {
                pokemon.copy(isFavorite = !pokemon.isFavorite)
            } else {
                pokemon
            }
        }
        viewModelScope.launch {
            if (_pokemonDetails.value != null) {
                val pokemon = _pokemonDetails.value!!.find { it.id == pokemonId }
                pokemon?.let {
                    if (it.isFavorite) {
                        favouritesRepository.insertItem(Favourite(it.id, it.name))
                    } else {
                        favouritesRepository.deleteItem(Favourite(it.id, it.name))
                    }
                }
            }
        }
    }
}
